@@ -15,7 +15,6 @@
 
 use std::error;
 use std::fmt;
-use std::fmt::Write;
 use std::str;
 
 use regex::Regex;
@@ -281,7 +280,7 @@ impl fmt::Display for ParseCodeIdError {
 ///  - **GO Build ID**: Nested GO build identifiers comprising `actionID/[.../]contentID`.
 #[derive(Clone, Default, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub struct CodeId {
-    inner: String,
+    inner: Vec<u8>,
 }
 
 impl CodeId {
@@ -290,21 +289,30 @@ impl CodeId {
         Self::default()
     }
 
-    /// Constructs a `CodeId` from its string representation.
-    pub fn new(mut string: String) -> Self {
-        string.make_ascii_lowercase();
-        CodeId { inner: string }
+    /// Constructs a `CodeId` from a binary buffer.
+    pub fn from_vec(vec: Vec<u8>) -> Self {
+        CodeId { inner: vec }
     }
 
     /// Constructs a `CodeId` from a binary slice.
-    pub fn from_binary(slice: &[u8]) -> Self {
-        let mut string = String::with_capacity(slice.len() * 2);
+    pub fn from_slice(slice: &[u8]) -> Self {
+        Self::from_vec(slice.into())
+    }
 
-        for byte in slice {
-            write!(&mut string, "{:02x}", byte).expect("");
+    /// Parses a `CodeId` from a hexadecimal string.
+    pub fn parse_hex(string: &str) -> Result<Self, ParseCodeIdError> {
+        if string.len() % 2 != 0 {
+            return Err(ParseCodeIdError);
         }
 
-        Self::new(string)
+        let vec = string
+            .as_bytes()
+            .chunks(2)
+            .map(|chunk| u8::from_str_radix(unsafe { str::from_utf8_unchecked(chunk) }, 16))
+            .collect::<Result<_, _>>()
+            .map_err(|_| ParseCodeIdError)?;
+
+        Ok(Self::from_vec(vec))
     }
 
     /// Returns whether this identifier is nil, i.e. it is empty.
@@ -312,15 +320,24 @@ impl CodeId {
         self.inner.is_empty()
     }
 
-    /// Returns the string representation of this code identifier.
-    pub fn as_str(&self) -> &str {
-        self.inner.as_str()
+    /// Returns the binary representation of this code identifier.
+    pub fn as_slice(&self) -> &[u8] {
+        &self.inner
+    }
+
+    /// If this `DebugId` holds a UUID, return it.
+    pub fn uuid(&self) -> Option<Uuid> {
+        Uuid::from_slice(&self.inner).ok()
     }
 }
 
 impl fmt::Display for CodeId {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.write_str(&self.inner)
+        for byte in &self.inner {
+            write!(f, "{:02x}", byte)?;
+        }
+
+        Ok(())
     }
 }
 
@@ -330,23 +347,23 @@ impl fmt::Debug for CodeId {
     }
 }
 
-impl From<String> for CodeId {
-    fn from(string: String) -> Self {
-        Self::new(string)
-    }
-}
-
-impl From<&'_ str> for CodeId {
-    fn from(string: &str) -> Self {
-        Self::new(string.into())
-    }
-}
-
 impl str::FromStr for CodeId {
     type Err = ParseCodeIdError;
 
     fn from_str(string: &str) -> Result<Self, ParseCodeIdError> {
-        Ok(Self::new(string.into()))
+        Self::parse_hex(string)
+    }
+}
+
+impl From<Vec<u8>> for CodeId {
+    fn from(vec: Vec<u8>) -> Self {
+        CodeId::from_vec(vec)
+    }
+}
+
+impl From<&'_ [u8]> for CodeId {
+    fn from(slice: &[u8]) -> Self {
+        CodeId::from_slice(slice)
     }
 }
 
@@ -359,14 +376,15 @@ mod serde_support {
 
     impl Serialize for CodeId {
         fn serialize<S: Serializer>(&self, serializer: S) -> Result<S::Ok, S::Error> {
-            serializer.serialize_str(self.as_str())
+            serializer.serialize_str(&self.to_string())
         }
     }
 
     impl<'de> Deserialize<'de> for CodeId {
         fn deserialize<D: Deserializer<'de>>(deserializer: D) -> Result<Self, D::Error> {
             let string = String::deserialize(deserializer)?;
-            Ok(CodeId::new(string))
+            CodeId::parse_hex(&string)
+                .map_err(|_| de::Error::invalid_value(Unexpected::Str(&string), &"CodeId"))
         }
     }
 
